#!/usr/bin/env python
# encoding: utf-8
"""
Eaeri.py

Created by Keoni on 2008-02-04.
Copyright (c) 2008 University of Wisconsin SSEC. All rights reserved.

    This file is part of the AERI ingest software.

    The AERI ingest software is free software; you can redistribute it and/or 
    modify it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""

import logging,os,win32,shutil,imp,time, string, re
import Config
from process import process
import logging

LOG = logging.getLogger( "Eaeri" )


# EAERI SUBSYSTEM OBJECT CLASSES
# These are implementations of the hardware proxy object interfaces done with ACQEAERI.
# Ideally these are drop-in replacements that the control logic can use.

class EaeriHousekeeping:
    "An adapter class implementing the Housekeeping interface using EAERI software."
    
    def __init__(self, ear):
        self._eaeri = ear
        # the current file that we're recording housekeeping to
        self._current_hk_fn = None
        # executable we can use
        self.acq = ear.acq
        LOG = logging.getLogger("EaeriHousekeeping")
                        
    def start(self): 
        """Bootstrap the housekeeping collection software.
        """                
        # this is already taken care of by Eaeri start()
        # FIXME: consider verifying that the housekeeping is being recorded and verify a record
        return self
        
    def peek(self):
        """Read a record from the high-resolution housekeeping to the screen
        """
        # How to find out the filename currently being written?
        #LOG.warning("EaeriHousekeeping.peek() not implemented")
        print "waiting 10 seconds for housekeeping data.. please wait."
        time.sleep(10)
        cm = '/cygdrive/c/config/channel.map'
        ec = '/cygdrive/c/config/eaeri_channels'
        fn = win32.windows_path(Config.calPath(time.strftime('%Y%m%d_%H%M%S.hk',time.gmtime())))
        self.acq('-k%s' % fn)

        fp = open(cm)
        txt = fp.readlines()
        fp.close()

        # Create dictionary of eaeri channels to AERI channels
        fp2 = open(ec)
        txt2 = fp2.readlines()
        fp2.close()

        g = {}
        for tmp in txt2:
           tmp = string.replace(tmp, '\n', '')
           if tmp != '':
             l, x = tmp.split('=')
             g[l] = x

        # Read last line of housekeeping output file
        fp3 = open(fn, 'r')
        txt3 = fp3.readlines()
        fp3.close()
        tl =  txt3[len(txt3)-1]
        t2 = string.split(tl, ',')
        h = {}
        for each in t2[1:]:
           k, v = each.split('=')
           h[k] = v

        for src in txt:
         if src[0][0] != '(':
           pattern = "([-0-9]+[' ']+[a-zA-Z0-9]+)"
           x = re.compile(pattern, re.IGNORECASE)
           e = x.findall(src)
           if e != []:
              ch_pattern = "([-0-9]+)"
              ch_name_pattern = "([a-zA-Z]+[a-zA-Z0-9]+)"
              y = re.compile(ch_pattern, re.IGNORECASE)
              z = re.compile(ch_name_pattern, re.IGNORECASE)
              ch = y.findall(e[0])
              ch_name = z.findall(e[0])[0]
              try:
                 ea_ch_name = g[ch_name]
              except:
                 ea_ch_name = "NA"
              try:
                 num = string.atof(h[ea_ch_name])
                 # formula for how many tabs
                 if len(ch_name) > 29:
                   print "channel %s: %5.4f\t: %s : %s" % (ch[0], num, ch_name, ea_ch_name)
                 elif len(ch_name) > 21:
                   print "channel %s: %5.4f\t: %s\t\t: %s" % (ch[0], num, ch_name, ea_ch_name)
                 elif len(ch_name) > 13:
                   print "channel %s: %5.4f\t: %s\t\t\t: %s" % (ch[0], num, ch_name, ea_ch_name)
                 elif len(ch_name) > 7:
                   print "channel %s: %5.4f\t: %s\t\t\t\t: %s" % (ch[0], num, ch_name, ea_ch_name)
                 else:
                   print "channel %s: %5.4f\t: %s\t\t\t: %s" % (ch[0], num, ch_name, ea_ch_name)
              except:
                 abc = 0
        return self
       
    def syncAndMarkStartOfSet(self, ifgTask=None):
        """ Synchronize the start of housekeeping collection to the start of interferogram acquisition.
        Start AERIHK recording the data for this acquisition.
        Return no sooner than the start of the interferogram acquire.
        FIXME: Currently does not use pipe synchronization to ACQ96 to find the start-of-scan.
        """
        # send housekeeping between scans down the drain for now
        drainfile = win32.windows_path(Config.calPath(time.strftime('%Y%m%d_%H%M%S.hk',time.gmtime())))
        LOG.debug('start-of-scan, drained housekeeping to ' + drainfile)
        self.acq('-k%s' % drainfile)
        return self
      
    def writeHousekeeping(self, pathname):
        """ Write the buffered housekeeping residing in AERIHK buffers to the pathname provided.
        FIXME: the .SET file gets read based on this name. is this a problem?
        raises EnvironmentError if any problem collecting housekeeping occurred
        """
        #hkfile = win32.windows_path(os.path.abspath(pathname))
        hkfile = pathname
        LOG.debug("writeHousekeeping to %s" % hkfile)
        self.acq('-k%s' % hkfile)
        # FIXME: other checks for errors?
        dirname, filename = win32.path_split(pathname)
        assert( os.path.exists(win32.unix_path(pathname)) )

    def kill(self):
        LOG.debug("EaeriHousekeeping kill")


class EaeriMirror:
    """An adapter class which uses ACQEAERI to command the mirror.
    """
    def __init__(self, ear):
        self._eaeri = ear
        # the current file that we're recording housekeeping to
        self._current_hk_fn = None
        # executable we can use
        self.acq = ear.acq
        self.LOG = logging.getLogger("EaeriMirror")
                        
    def start(self):
        """Start up mirror daemon - this shared memory server holds the state of the mirror system
        """
        return self # as a task, effectively
        
    def kill(self):
        pass        
        
    def isReady(self):
        """return boolean, whether or not the mirror is available to rotate
        """
        raise NotImplemented
        return True #FIXME

    def getState(self):
        """ return MIR_* constant indicating the state of the mirror
        """
        raise NotImplemented
        
    def programMirrorController(self):
        self.LOG.debug("programMirrorController not implemented")
        raise NotImplemented
        
    def rotateToPosition(self,mirror_position,schedule_pathname=None):
        """rotate the mirror go a given scene using the default mirror schedule file in the workdir
        Note that we're not using aerimirc.exe here, which itself is part mirror / part sequencer. 
        This means that the sequencer takes the on the responsibility (and rightfully so) of 
        sequencing the mirror schedule file.
        """
        # open mirror schedule
        # get angle for named position
        # send command to go to angle
        self.LOG.info("rotating to position %s" % mirror_position)
        #FIXME: determine whether it actually made it there, or stalled out
        raise NotImplemented
        
    def requestFreeFromSafing(self):
        """ Signal the mirror system that now is a good time to release the mirror from safing.
        In the case that the mirror isn't safed, no-op.
        If the mirror cannot be released from safe, no-op.
        """
        self.LOG.debug('requestFreeFromSafing not implemented')
        raise NotImplemented
    
    def forceReady(self):
        """Force the mirror to a homed and ready state.
        """
        self.LOG.debug("forceReady not implemented")
        raise NotImplemented
        
    def recordPosition(self,pos_pathname):
        """ write a mirror position file to the given pathname.
        """
        if os.path.exists( pos_pathname ):
            os.unlink(pos_pathname)
        # request current position
        
        # FIXME - how to do an encoder position request and translate to angle?
        self.LOG.debug("encoder position readout not implemented")

        # return the encoder position, or None if no information was obtained
        if not os.path.exists( pos_pathname ):            
            self.LOG.warning("no position information from EAERI mirror")
            return None
        try:        
            return int( file(pos_pathname,'rt').read()[1:] )
        except:
            LOG.warning("no position report from mirror")
            return None

class EaeriAutonomic:
    def __init__(self, ear):
        self._eaeri = ear
        pass                            

    def setHatch(self,state=True):
        """Change the hatch-enable state on the hatch daemon
        """
        # from socket import socket
        # s = socket()
        # log.debug('connecting to tcp port %d' % self.hatch_tcp)
        # s.connect( ('localhost',self.hatch_tcp) )
        # s.send( state and 'enable\n' or 'disable\n' )
        # s.close()
        if state: state = '1'
        else: state = '0'
        self._eaeri.acq('-k',state)
    
    def startHatch(self):
        #env = os.environ.copy()
        #mypid = os.spawnvpe( os.P_NOWAIT, 'python', [ Config.HATCH_EXE ], env )

        #mypid = os.fork()
        #if not mypid: 
        #    for k,v in env.items:
        #        os.environ[k] = v
        #    os.execvp( 'python', [ Config.HATCH_EXE ] )
        #self.hatch = process( 'hatchd', pid = mypid )
        # self.hatch.start()
        
        # return self.hatch
        raise NotImplemented
        
    def startMirrorSafing(self):
        """bootstrap mirror safing daemon; return task representing it
        """
        # self.mirsaf.start()
        return self.mirsaf
        
    def start(self):
        # self.mirsaf.start()
        # self.hatch.start()
        return self
        
    def preheatMirrorMotor(self):
        """preheat the mirror motor. return when the mirror is warm enough to rotate.
        """
        # env = { 'MIRROR_SCD': Config.MIRROR_BEG, 'AEMIRSAF_CFG': Config.MOTHEAT_CFG, 'CHANNEL_MAP': Config.CHANNEL_MAP }
        # p = process(Config.AEMIRSAF_EXE, 
        #             environ = env,
        #             output_stream=Config.logStream("preheat.out"),
        #             output_prefix='aemirsaf')
        # log.info("launching preheat")
        # p()
        raise NotImplemented
        
    def kill(self):
        pass

class eaeri_acquire_task:
    def __init__(self,proc,fn_log,flog,hax):
        self._proc = proc
        self._fnlog = fn_log
        self._flog = flog
        self._hax = hax
        
    def wait(self):
        self._proc.wait()
        # Move Files to Temp directory
        self._hax()
        # read the log file and cat it to our log
        try:
            self._flog.write( file(self._fnlog,'rt').read() )
        except:
            LOG.warning("unable to transcribe ACQ96NT.LOG to log file")    
        try:
            os.unlink(self._fnlog)
        except:
            LOG.warning("unable to remove old ACQ96NT.LOG")
            
    def kill(self):
        self._proc.kill()

class EaeriInterferometer:
    def __init__(self,ear):
        self._eaeri = ear
        self._acqlog = Config.logStream('acq96.log')

    def startIngestSpectra(self,set_pathname):
        """Fork a process to ingest the a given .SET file pathname, returning a task proxy.
        Generates SPC data in the same directory as the SET file.
        """
        dirname, filename = win32.path_split( set_pathname )
        stem,exten = os.path.splitext( filename )
        import acqeaeri_hax, subprocess
        hax = acqeaeri_hax.acqeaeri_file_hax( Config.ACQEAERI_EXE, win32.unix_path(dirname) )
        acq96 = self._eaeri.acq_start('M', stem, work_dir=dirname)
        from Interferometer import acquire_task # FUTURE: clean this up
        return eaeri_acquire_task(acq96,win32.path_join(dirname,Config.ACQ96_LOG),self._acqlog, hax)


class Eaeri:
    """ Wrapper class for daemons representing the Eaeri services for housekeeping et al
    This object is also a factory for subsystem objects (housekeeping, mirror, etc) with identical interfaces
    It also coordinates any race conditions or locks that need to happen in order to prevent clobberings
    """
    def __init__(self, port = None, simulate = False):
        
        self._acq = process(Config.ACQEAERI_EXE,
                         output_stream=Config.logStream('acqeaeri.out'),
                         output_prefix='acqeaeri' )
                         
        if simulate:
            self._sim = process(Config.EAERI_SIMULATOR_EXE,
                         output_stream=Config.logStream('eaerisim.out'),
                         output_prefix='eaerisim' )
        else:
            self._sim = None

        # read pythonic configuration file for EAERI
        fn = Config.EAERI_CFG
        nfo = file(fn,'U'), fn, ('.py', 'U', 1)
        mod = imp.load_module('eaeri_cfg', *nfo)
        self._cfg = mod
        
        self._simulate = simulate
        
        # connection port override - currently not being used (2008.05)
        if port is not None:
            self._cfg.port = port
        if 'port' not in dir(self._cfg):
            if simulate:
                self._cfg.port = Config.EAERI_DEFAULT_SIMULATOR_PORT 
            else:
                self._cfg.port = Config.EAERI_DEFAULT_PORT        
        
    def start(self):
        """ bootstrap system daemons needed for EAERI hardware
        """
        if self._simulate:
            LOG.info("starting simulator")
            self._sim.start()
        
        # return a kill()-able corresponding to what was started
        return self 
    
    def acq(self,*args,**kwargs):
        """ invoke ACQEAERI with default parameters and provided additional parameters
        """
        #  '-r%d' % self._cfg.port,
        return self._acq( *args, **kwargs)

    def acq_start(self,*args,**kwargs):
        """ invoke ACQEAERI with default parameters and provided additional parameters
        """
        # '-r%d' % self._cfg.port,
        self._acq.start( *args, **kwargs)
        return self._acq
    
    def subsys_interferometer(self):
        return EaeriInterferometer(self)
                    
    def subsys_housekeeping(self):
        return EaeriHousekeeping(self)
        
    def subsys_mirror(self):
        return EaeriMirror(self)

    def subsys_hatch(self):
        return EaeriMirror(self)
        
    def subsys_interferometer(self):
        return EaeriInterferometer(self)
    
def test_hk_simple(args = None):
    from Housekeeping import test_simple
    return test_simple(args, Eaeri().subsys_housekeeping)

def test_hk_record(args = None):
    from Housekeeping import test_record
    return test_record(args, Eaeri().subsys_housekeeping)

def test_interfer(args = None):
    from Interferometer import test_interfer as mytest
    mytest(args, Eaeri().subsys_interferometer)

if __name__=='__main__':
    logging.basicConfig()
    LOG.setLevel(logging.DEBUG)
    test_interfer()
    test_hk_simple()
    test_hk_record()
    
